@@ -1,8 +1,9 @@
 import operator
 
-from .compat import PY2
-from .compat import PY3
+from .compat import string_types
 from .compat import with_metaclass
+from .utils import await_
+from .utils import identity
 
 
 class _ProxyMethods(object):
@@ -61,11 +62,35 @@ class _ProxyMetaType(type):
 
 
 class Proxy(with_metaclass(_ProxyMetaType)):
+    """
+    A proxy implementation in pure Python, using slots. You can subclass this to add
+    local methods or attributes, or enable __dict__.
+
+    The most important internals:
+
+    * ``__factory__`` is the callback that "materializes" the object we proxy to.
+    * ``__target__`` will contain the object we proxy to, once it's "materialized".
+    * ``__resolved__`` is a boolean, `True` if factory was called.
+    * ``__wrapped__`` is a property that does either:
+
+      * return ``__target__`` if it's set.
+      * calls ``__factory__``, saves result to ``__target__`` and returns said result.
+    """
+
     __slots__ = '__target__', '__factory__'
 
     def __init__(self, factory):
         object.__setattr__(self, '__factory__', factory)
 
+    @property
+    def __resolved__(self, __getattr__=object.__getattribute__):
+        try:
+            __getattr__(self, '__target__')
+        except AttributeError:
+            return False
+        else:
+            return True
+
     @property
     def __wrapped__(self, __getattr__=object.__getattribute__, __setattr__=object.__setattr__, __delattr__=object.__delattr__):
         try:
@@ -99,8 +124,8 @@ class Proxy(with_metaclass(_ProxyMetaType)):
     def __class__(self):
         return self.__wrapped__.__class__
 
-    @__class__.setter
-    def __class__(self, value):
+    @__class__.setter  # noqa: F811
+    def __class__(self, value):  # noqa: F811
         self.__wrapped__.__class__ = value
 
     @property
@@ -117,22 +142,35 @@ class Proxy(with_metaclass(_ProxyMetaType)):
     def __str__(self):
         return str(self.__wrapped__)
 
-    if PY3:
-        def __bytes__(self):
-            return bytes(self.__wrapped__)
+    def __bytes__(self):
+        return bytes(self.__wrapped__)
 
-    def __repr__(self):
-        return '<%s at 0x%x for %s at 0x%x>' % (
-            type(self).__name__, id(self),
-            type(self.__wrapped__).__name__,
-            id(self.__wrapped__))
+    def __repr__(self, __getattr__=object.__getattribute__):
+        try:
+            target = __getattr__(self, '__target__')
+        except AttributeError:
+            return '<{} at 0x{:x} with factory {!r}>'.format(type(self).__name__, id(self), self.__factory__)
+        else:
+            return '<{} at 0x{:x} wrapping {!r} at 0x{:x} with factory {!r}>'.format(
+                type(self).__name__, id(self), target, id(target), self.__factory__
+            )
+
+    def __fspath__(self):
+        wrapped = self.__wrapped__
+        if isinstance(wrapped, string_types):
+            return wrapped
+        else:
+            fspath = getattr(wrapped, '__fspath__', None)
+            if fspath is None:
+                return wrapped
+            else:
+                return fspath()
 
     def __reversed__(self):
         return reversed(self.__wrapped__)
 
-    if PY3:
-        def __round__(self):
-            return round(self.__wrapped__)
+    def __round__(self):
+        return round(self.__wrapped__)
 
     def __lt__(self, other):
         return self.__wrapped__ < other
@@ -168,7 +206,10 @@ class Proxy(with_metaclass(_ProxyMetaType)):
             setattr(self.__wrapped__, name, value)
 
     def __getattr__(self, name):
-        return getattr(self.__wrapped__, name)
+        if name in ('__wrapped__', '__factory__'):
+            raise AttributeError(name)
+        else:
+            return getattr(self.__wrapped__, name)
 
     def __delattr__(self, name, __delattr__=object.__delattr__):
         if hasattr(type(self), name):
@@ -185,8 +226,8 @@ class Proxy(with_metaclass(_ProxyMetaType)):
     def __mul__(self, other):
         return self.__wrapped__ * other
 
-    def __div__(self, other):
-        return operator.div(self.__wrapped__, other)
+    def __matmul__(self, other):
+        return self.__wrapped__ @ other
 
     def __truediv__(self, other):
         return operator.truediv(self.__wrapped__, other)
@@ -195,7 +236,7 @@ class Proxy(with_metaclass(_ProxyMetaType)):
         return self.__wrapped__ // other
 
     def __mod__(self, other):
-        return self.__wrapped__ ^ other
+        return self.__wrapped__ % other
 
     def __divmod__(self, other):
         return divmod(self.__wrapped__, other)
@@ -227,6 +268,9 @@ class Proxy(with_metaclass(_ProxyMetaType)):
     def __rmul__(self, other):
         return other * self.__wrapped__
 
+    def __rmatmul__(self, other):
+        return other @ self.__wrapped__
+
     def __rdiv__(self, other):
         return operator.div(other, self.__wrapped__)
 
@@ -272,8 +316,8 @@ class Proxy(with_metaclass(_ProxyMetaType)):
         self.__wrapped__ *= other
         return self
 
-    def __idiv__(self, other):
-        self.__wrapped__ = operator.idiv(self.__wrapped__, other)
+    def __imatmul__(self, other):
+        self.__wrapped__ @= other
         return self
 
     def __itruediv__(self, other):
@@ -327,10 +371,6 @@ class Proxy(with_metaclass(_ProxyMetaType)):
     def __int__(self):
         return int(self.__wrapped__)
 
-    if PY2:
-        def __long__(self):
-            return long(self.__wrapped__)  # flake8: noqa
-
     def __float__(self):
         return float(self.__wrapped__)
 
@@ -341,7 +381,10 @@ class Proxy(with_metaclass(_ProxyMetaType)):
         return hex(self.__wrapped__)
 
     def __index__(self):
-        return operator.index(self.__wrapped__)
+        if hasattr(self.__wrapped__, '__index__'):
+            return operator.index(self.__wrapped__)
+        else:
+            return int(self.__wrapped__)
 
     def __len__(self):
         return len(self.__wrapped__)
@@ -376,5 +419,23 @@ class Proxy(with_metaclass(_ProxyMetaType)):
     def __iter__(self):
         return iter(self.__wrapped__)
 
+    def __next__(self):
+        return next(self.__wrapped__)
+
     def __call__(self, *args, **kwargs):
         return self.__wrapped__(*args, **kwargs)
+
+    def __reduce__(self):
+        return identity, (self.__wrapped__,)
+
+    def __reduce_ex__(self, protocol):
+        return identity, (self.__wrapped__,)
+
+    if await_:
+        from .utils import __aenter__
+        from .utils import __aexit__
+        from .utils import __aiter__
+        from .utils import __anext__
+        from .utils import __await__
+
+        __aiter__, __anext__, __await__, __aenter__, __aexit__  # noqa
